import numpy as np
import pytest

from albumentations.augmentations.bbox_utils import normalize_bbox, denormalize_bbox, normalize_bboxes, \
    denormalize_bboxes, calculate_bbox_area, filter_bboxes_by_visibility, convert_bbox_to_albumentations,\
    convert_bbox_from_albumentations, convert_bboxes_to_albumentations, convert_bboxes_from_albumentations
from albumentations.core.composition import ComposeWithBoxes
from albumentations.core.transforms_interface import NoOp
from albumentations.augmentations.transforms import RandomSizedCrop


@pytest.mark.parametrize(['bbox', 'expected'], [
    [[15, 25, 100, 200], [0.0375, 0.125, 0.25, 1.0]],
    [[15, 25, 100, 200, 99], [0.0375, 0.125, 0.25, 1.0, 99]],
])
def test_normalize_bbox(bbox, expected):
    normalized_bbox = normalize_bbox(bbox, 200, 400)
    assert normalized_bbox == expected


@pytest.mark.parametrize(['bbox', 'expected'], [
    [[0.0375, 0.125, 0.25, 1.0], [15, 25, 100, 200]],
    [[0.0375, 0.125, 0.25, 1.0, 99], [15, 25, 100, 200, 99]],
])
def test_denormalize_bbox(bbox, expected):
    denormalized_bbox = denormalize_bbox(bbox, 200, 400)
    assert denormalized_bbox == expected


@pytest.mark.parametrize('bbox', [[15, 25, 100, 200], [15, 25, 100, 200, 99]])
def test_normalize_denormalize_bbox(bbox):
    normalized_bbox = normalize_bbox(bbox, 200, 400)
    denormalized_bbox = denormalize_bbox(normalized_bbox, 200, 400)
    assert denormalized_bbox == bbox


@pytest.mark.parametrize('bbox', [[0.0375, 0.125, 0.25, 1.0], [0.0375, 0.125, 0.25, 1.0, 99]])
def test_denormalize_normalize_bbox(bbox):
    denormalized_bbox = denormalize_bbox(bbox, 200, 400)
    normalized_bbox = normalize_bbox(denormalized_bbox, 200, 400)
    assert normalized_bbox == bbox


def test_normalize_bboxes():
    bboxes = [[15, 25, 100, 200], [15, 25, 100, 200, 99]]
    normalized_bboxes_1 = normalize_bboxes(bboxes, 200, 400)
    normalized_bboxes_2 = [normalize_bbox(bboxes[0], 200, 400), normalize_bbox(bboxes[1], 200, 400)]
    assert normalized_bboxes_1 == normalized_bboxes_2


def test_denormalize_bboxes():
    bboxes = [[0.0375, 0.125, 0.25, 1.0], [0.0375, 0.125, 0.25, 1.0, 99]]
    denormalized_bboxes_1 = denormalize_bboxes(bboxes, 200, 400)
    denormalized_bboxes_2 = [denormalize_bbox(bboxes[0], 200, 400), denormalize_bbox(bboxes[1], 200, 400)]
    assert denormalized_bboxes_1 == denormalized_bboxes_2


@pytest.mark.parametrize(['bbox', 'rows', 'cols', 'expected'], [
    [[0, 0, 1, 1], 50, 100, 5000],
    [[0.2, 0.2, 1, 1, 99], 50, 50, 1600],
])
def test_calculate_bbox_area(bbox, rows, cols, expected):
    area = calculate_bbox_area(bbox, rows, cols)
    assert area == expected


@pytest.mark.parametrize(['bbox', 'source_format', 'expected'], [
    [[20, 30, 40, 50], 'coco', [0.2, 0.3, 0.6, 0.8]],
    [[20, 30, 40, 50, 99], 'coco', [0.2, 0.3, 0.6, 0.8, 99]],
    [[20, 30, 60, 80], 'pascal_voc', [0.2, 0.3, 0.6, 0.8]],
    [[20, 30, 60, 80, 99], 'pascal_voc', [0.2, 0.3, 0.6, 0.8, 99]],
])
def test_convert_bbox_to_albumentations(bbox, source_format, expected):
    image = np.ones((100, 100, 3))

    converted_bbox = convert_bbox_to_albumentations(bbox, rows=image.shape[0], cols=image.shape[1],
                                                    source_format=source_format)
    assert converted_bbox == expected


@pytest.mark.parametrize(['bbox', 'target_format', 'expected'], [
    [[0.2, 0.3, 0.6, 0.8], 'coco', [20, 30, 40, 50]],
    [[0.2, 0.3, 0.6, 0.8, 99], 'coco', [20, 30, 40, 50, 99]],
    [[0.2, 0.3, 0.6, 0.8], 'pascal_voc', [20, 30, 60, 80]],
    [[0.2, 0.3, 0.6, 0.8, 99], 'pascal_voc', [20, 30, 60, 80, 99]],
])
def test_convert_bbox_from_albumentations(bbox, target_format, expected):
    image = np.ones((100, 100, 3))
    converted_bbox = convert_bbox_from_albumentations(bbox, rows=image.shape[0], cols=image.shape[1],
                                                      target_format=target_format)
    assert converted_bbox == expected


@pytest.mark.parametrize(['bbox', 'bbox_format'], [
    [[20, 30, 40, 50], 'coco'],
    [[20, 30, 40, 50, 99], 'coco'],
    [[20, 30, 60, 80], 'pascal_voc'],
    [[20, 30, 60, 80, 99], 'pascal_voc'],
])
def test_convert_bbox_to_albumentations_and_back(bbox, bbox_format):
    image = np.ones((100, 100, 3))
    converted_bbox = convert_bbox_to_albumentations(bbox, rows=image.shape[0], cols=image.shape[1],
                                                    source_format=bbox_format)
    converted_back_bbox = convert_bbox_from_albumentations(converted_bbox, rows=image.shape[0], cols=image.shape[1],
                                                           target_format=bbox_format)
    assert converted_back_bbox == bbox


def test_convert_bboxes_to_albumentations():
    bboxes = [[20, 30, 40, 50], [30, 40, 50, 60, 99]]
    image = np.ones((100, 100, 3))
    converted_bboxes = convert_bboxes_to_albumentations(bboxes, rows=image.shape[0], cols=image.shape[1],
                                                        source_format='coco')
    converted_bbox_1 = convert_bbox_to_albumentations(bboxes[0], rows=image.shape[0], cols=image.shape[1],
                                                      source_format='coco')
    converted_bbox_2 = convert_bbox_to_albumentations(bboxes[1], rows=image.shape[0], cols=image.shape[1],
                                                      source_format='coco')
    assert converted_bboxes == [converted_bbox_1, converted_bbox_2]


def test_convert_bboxes_from_albumentations():
    bboxes = [[0.2, 0.3, 0.6, 0.8], [0.3, 0.4, 0.7, 0.9, 99]]
    image = np.ones((100, 100, 3))
    converted_bboxes = convert_bboxes_to_albumentations(bboxes, rows=image.shape[0], cols=image.shape[1],
                                                        source_format='coco')
    converted_bbox_1 = convert_bbox_to_albumentations(bboxes[0], rows=image.shape[0], cols=image.shape[1],
                                                      source_format='coco')
    converted_bbox_2 = convert_bbox_to_albumentations(bboxes[1], rows=image.shape[0], cols=image.shape[1],
                                                      source_format='coco')
    assert converted_bboxes == [converted_bbox_1, converted_bbox_2]


@pytest.mark.parametrize(['bboxes', 'bbox_format'], [
    [[[20, 30, 40, 50]], 'coco'],
    [[[20, 30, 40, 50, 99], [10, 40, 30, 20, 9]], 'coco'],
    [[[20, 30, 60, 80]], 'pascal_voc'],
    [[[20, 30, 60, 80, 99]], 'pascal_voc'],
])
def test_compose_with_bbox_noop(bboxes, bbox_format):
    image = np.ones((100, 100, 3))
    aug = ComposeWithBoxes([NoOp(p=1.)], bbox_format)
    transformed = aug(image=image, bboxes=bboxes)
    assert np.array_equal(transformed['image'], image)
    assert transformed['bboxes'] == bboxes


@pytest.mark.parametrize(['bboxes', 'bbox_format', 'labels'], [
    [[[20, 30, 60, 80]], 'pascal_voc', {'label': [1]}],
    [[[20, 30, 60, 80]], 'pascal_voc', {'id': [3]}],
    [[[20, 30, 60, 80], [30, 40, 40, 50]], 'pascal_voc', {'id': [3, 1]}],
])
def test_compose_with_bbox_noop_label_outside(bboxes, bbox_format, labels):
    image = np.ones((100, 100, 3))
    aug = ComposeWithBoxes([NoOp(p=1.)], bbox_format, label_fields=list(labels.keys()))
    transformed = aug(image=image, bboxes=bboxes, **labels)
    assert np.array_equal(transformed['image'], image)
    assert transformed['bboxes'] == bboxes
    for k, v in labels.items():
        assert transformed[k] == v


def test_random_sized_crop_size():
    image = np.ones((100, 100, 3))
    bboxes = [[0.2, 0.3, 0.6, 0.8], [0.3, 0.4, 0.7, 0.9, 99]]
    aug = RandomSizedCrop((70, 90), 50, 50, p=1.)
    transformed = aug(image=image, bboxes=bboxes)
    assert transformed['image'].shape == (50, 50, 3)
    assert len(bboxes) == len(transformed['bboxes'])



from __future__ import division

import random

import numpy as np
from albumentations.augmentations.bbox_utils import convert_bboxes_from_albumentations, \
    convert_bboxes_to_albumentations, filter_bboxes


__all__ = ['Compose', 'OneOf', 'OneOrOther', 'ComposeWithBoxes']


class Compose(object):
    """Compose transforms together.

    Args:
        transforms (list): list of transformations to compose.
        preprocessing_transforms (list): list of transforms to run before transforms
        p (float): probability of applying all list of transforms. Default: 1.0.
    """

    def __init__(self, transforms, preprocessing_transforms=[], p=1.0):
        self.preprocessing_transforms = preprocessing_transforms
        self.transforms = [t for t in transforms if t is not None]
        self.p = p

    def __call__(self, **data):
        need_to_run = random.random() < self.p
        return self.run_transforms_if_needed(need_to_run, data)

    def run_transforms_if_needed(self, need_to_run, data):
        for t in self.preprocessing_transforms:
            data = t(**data)
        if need_to_run:
            for t in self.transforms:
                data = t(**data)
        return data


class ComposeWithBoxes(Compose):
    """Compose transforms and handle all transformations regrading bounding boxes

    Args:
        transforms (list): list of transformations to compose.
        preprocessing_transforms (list): list of transforms to run before transforms
        bbox_format (str): format of bounding boxes. Should be 'coco' or 'pascal_voc'.
        label_fields (list): list of fields that are joined with boxes, e.g labels. Should be same type as boxes.
        min_area (float): minimum area of a bounding box. All bounding boxes whose visible area in pixels
            is less than this value will be removed. Default: 0.0.
        min_visibility (float): minimum fraction of area for a bounding box to remain this box in list. Default: 0.0.
        p (float): probability of applying all list of transforms. Default: 1.0.
    """

    def __init__(self, transforms, bbox_format, label_fields=[], min_area=0., min_visibility=0.,
                 preprocessing_transforms=[], p=1.0):
        super(ComposeWithBoxes, self).__init__(transforms, preprocessing_transforms, p=p)
        self.bbox_format = bbox_format
        self.label_fields = label_fields
        self.min_area = min_area
        self.min_visibility = min_visibility

    def __call__(self, **data):
        need_to_run = random.random() < self.p
        if self.preprocessing_transforms or need_to_run:
            if 'bboxes' not in data:
                raise Exception('Please name field with bounding boxes `bboxes`')
            if self.label_fields:
                for field in self.label_fields:
                    bboxes_with_added_field = []
                    for bbox, field_value in zip(data['bboxes'], data[field]):
                        bboxes_with_added_field.append(list(bbox) + [field_value])
                    data['bboxes'] = bboxes_with_added_field

            rows, cols = data['image'].shape[:2]
            data['bboxes'] = convert_bboxes_to_albumentations(data['bboxes'], self.bbox_format, rows, cols,
                                                              check_validity=True)

            data = self.run_transforms_if_needed(need_to_run, data)

            rows, cols = data['image'].shape[:2]
            data['bboxes'] = filter_bboxes(data['bboxes'], rows, cols, self.min_area, self.min_visibility)

            data['bboxes'] = convert_bboxes_from_albumentations(data['bboxes'], self.bbox_format, rows, cols,
                                                                check_validity=True)

            if self.label_fields:
                for idx, field in enumerate(self.label_fields):
                    field_values = []
                    for bbox in data['bboxes']:
                        field_values.append(bbox[4 + idx])
                    data[field] = field_values
                data['bboxes'] = [bbox[:4] for bbox in data['bboxes']]
        return data


class OneOf(object):
    """Select on of transforms to apply

    Args:
        transforms (list): list of transformations to compose.
        p (float): probability of applying selected transform. Default: 0.5.
    """

    def __init__(self, transforms, p=0.5):
        self.transforms = transforms
        self.p = p
        transforms_ps = [t.p for t in transforms]
        s = sum(transforms_ps)
        self.transforms_ps = [t / s for t in transforms_ps]

    def __call__(self, **data):
        if random.random() < self.p:
            random_state = np.random.RandomState(random.randint(0, 2 ** 32 - 1))
            t = random_state.choice(self.transforms, p=self.transforms_ps)
            t.p = 1.
            data = t(**data)
        return data


class OneOrOther(object):
    def __init__(self, first, second, p=0.5):
        self.first = first
        first.p = 1.
        self.second = second
        second.p = 1.
        self.p = p

    def __call__(self, **data):
        return self.first(**data) if random.random() < self.p else self.second(**data)
